"""
Builder.py, module definition of Builder class.
This is a class designed for controlling the building phase.
"""
# from __future__ import absolute_import
# from __future__ import division
# from __future__ import print_function
# from __future__ import unicode_literals
# Copyright (C) 2015  Stefano Zaghi
#
# This file is part of FoBiS.py.
#
# FoBiS.py is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# FoBiS.py is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with FoBiS.py. If not, see <http://www.gnu.org/licenses/>.
# from future import standard_library
# standard_library.install_aliases()
# from builtins import str
# from builtins import range
# from builtins import *
# from builtins import object
try:
  from multiprocessing import Pool
  __parallel__ = True
except ImportError:
  print("Module 'multiprocessing' not found: parallel compilation disabled")
  __parallel__ = False
import operator
import os
from .Compiler import Compiler
from .utils import check_results, print_fake, syswork, safe_mkdir


class Builder(object):
  """Builder is an object that handles the building system, its attributes and methods."""

  def __init__(self, cliargs, print_n=None, print_w=None):
    """
    Parameters
    ----------
    cliargs : argparse object
    print_n : {None}
      function for printing normal message
    print_w : {None}
      function for printing emphized warning message

    Attributes
    ----------
    compiler : Compiler object
      compiler used
    src_dir = {"./"}
      directory into which search source files
    build_dir : {"./"}
      directory containing built files
    obj_dir : {"./"}
      directory containing compiled object files
    mod_dir : {"./"}
      directory containing .mod files
    lib_dir : {None}
      list of directories searched for libraries
    dinc : {None}
      list of directories for searching included files
    libs : {None}
      list of external libraries that are not into the path: must be passed with full paths
    vlibs : {None}
      list of external libraries that are not into the path and that are volatile (can changed thus triggering re-building): must be passed with full paths
    ext_libs : {None}
      list of external libraries that are into the path: are linked as '-llibrary.[a|so]' and searched into '-Ldir'
    ext_vlibs : {None}
      list of external libraries that are into the path and that are volatile (can changed thus triggering re-building): are linked as '-llibrary.[a|so]' and searched into '-Ldir'
    preprocessor : {None}
      activate preprocessor
    preprocessor_dir : {None}
      directory containing sources processed by preprocessor (if any); by default are removed after used
    preprocessor_ext : {None}
      list of file extensions to be processed by preprocessor; by default all files are preprocessed if preprocessor is used
    print_n : {None}
      function for printing normal message
    print_w : {None}
      function for printing emphized warning message
    jobs : {1}
      concurrent compiling jobs
    """

    if print_n is None:
      self.print_n = print_fake
    else:
      self.print_n = print_n

    if print_w is None:
      self.print_w = print_fake
    else:
      self.print_w = print_w

    self.compiler = Compiler(cliargs=cliargs, print_w=self.print_w)

    self.jobs = cliargs.jobs
    self._sanitize_dirs(src_dir=cliargs.src, build_dir=cliargs.build_dir, obj_dir=cliargs.obj_dir, mod_dir=cliargs.mod_dir, lib_dir=cliargs.lib_dir, dinc=cliargs.include)
    self._sanitize_files(libs=cliargs.libs, vlibs=cliargs.vlibs, ext_libs=cliargs.ext_libs, ext_vlibs=cliargs.ext_vlibs)
    self._set_preprocessor(preprocessor=cliargs.preprocessor,
                           preprocessor_dir=cliargs.preprocessor_dir,
                           preprocessor_ext=cliargs.preprocessor_ext,
                           preprocessor_no_o=cliargs.preprocessor_no_o,
                           preprocessor_args=cliargs.preprocessor_args)

    self.ar = cliargs.ar
    self.arflags = cliargs.arflags
    self.ranlib = cliargs.ranlib

    self.cmd_comp = self.compiler.compile_cmd(mod_dir=self.mod_dir)
    self.cmd_link = self.compiler.link_cmd(mod_dir=self.mod_dir)
    return

  @staticmethod
  def get_fc(cliargs):
    """
    Method for getting the compiler command built accordingly to the cli arguments.

    Parameters
    ----------
    cliargs : argparse object
    """
    return Compiler(cliargs=cliargs).fcs

  @staticmethod
  def get_cflags(cliargs):
    """
    Method for getting the compiling flags built accordingly to the cli arguments.

    Parameters
    ----------
    cliargs : argparse object
    """
    return Compiler(cliargs=cliargs).cflags

  @staticmethod
  def get_lflags(cliargs):
    """
    Method for getting the linking flags built accordingly to the cli arguments.

    Parameters
    ----------
    cliargs : argparse object
    """
    return Compiler(cliargs=cliargs).lflags

  @staticmethod
  def get_modsw(cliargs):
    """
    Method for getting the compiler modules switch built accordingly to the cli arguments.

    Parameters
    ----------
    cliargs : argparse object
    """
    # modsw = cliargs.modsw
    # if cliargs.compiler.lower() in Builder.s_compilers and modsw == '':
    #   modsw = Builder.s_compilers[cliargs.compiler.lower()][2]
    # return modsw
    return Compiler(compiler=cliargs.compiler, modsw=cliargs.modsw).modsw

  def _sanitize_dirs(self, src_dir, build_dir, obj_dir, mod_dir, lib_dir, dinc):
    """
    Method for sanitizing directory paths.

    Parameters
    ----------
    src_dir : list
      directory into which search source files
    build_dir : str
      directory containing built files
    obj_dir : str
      directory containing compiled object files
    mod_dir : str
      directory containing .mod files
    lib_dir : list
      list of directories searched for libraries
    dinc : list
      list of directories for searching included files
    """
    self.src_dir = ' '.join(src_dir)

    self.build_dir = os.path.normpath(build_dir)
    safe_mkdir(directory=self.build_dir, print_w=self.print_w)

    self.obj_dir = os.path.normpath(os.path.join(build_dir, obj_dir))
    safe_mkdir(directory=self.obj_dir, print_w=self.print_w)

    self.mod_dir = os.path.normpath(os.path.join(build_dir, mod_dir))
    safe_mkdir(directory=self.mod_dir, print_w=self.print_w)

    if lib_dir is None:
      self.lib_dir = []
    else:
      self.lib_dir = lib_dir
      for directory in self.lib_dir:
        directory = os.path.normpath(directory)
    if dinc is None:
      self.dinc = []
    else:
      self.dinc = dinc
      for directory in self.dinc:
        directory = os.path.normpath(directory)
    return

  def _sanitize_files(self, libs, vlibs, ext_libs, ext_vlibs):
    """
    Method for sanitizing files paths.

    Parameters
    ----------
    libs : list
      list of external libraries that are not into the path: must be passed with full paths
    vlibs : list
      list of external libraries that are not into the path and that are volatile (can changed thus triggering re-building): must be passed with full paths
    ext_libs : list
      list of external libraries that are into the path: are linked as '-llibrary.[a|so]' and searched into '-Ldir'
    ext_vlibs : list
      list of external libraries that are into the path and that are volatile (can changed thus triggering re-building): are linked as '-llibrary.[a|so]' and searched into '-Ldir'
    """
    if libs is None:
      self.libs = []
    else:
      self.libs = libs
      for lib in self.libs:
        lib = os.path.normpath(lib)
    if vlibs is None:
      self.vlibs = []
    else:
      self.vlibs = vlibs
      for lib in self.vlibs:
        lib = os.path.normpath(lib)
    if ext_libs is None:
      self.ext_libs = []
    else:
      self.ext_libs = ext_libs
      for lib in self.ext_libs:
        lib = os.path.normpath(lib)
    if ext_vlibs is None:
      self.ext_vlibs = []
    else:
      self.ext_vlibs = ext_vlibs
      for lib in self.ext_vlibs:
        lib = os.path.normpath(lib)
    return

  def _set_preprocessor(self, preprocessor, preprocessor_dir, preprocessor_ext,
                        preprocessor_no_o, preprocessor_args):
    """
    Set preprocessor data.

    Parameters
    ----------
    preprocessor : bool
      activate preprocessor
    preprocessor_dir : str
      directory containing sources processed by preprocessor; by default are removed after used
    preprocessor_ext : list
      list of file extensions to be processed by preprocessor; by default all files are preprocessed if preprocessor is used
    """
    self.preprocessor = preprocessor
    if preprocessor_ext is None:
      self.preprocessor_ext = []
    else:
      self.preprocessor_ext = preprocessor_ext
    self.preprocessor_dir = preprocessor_dir
    if self.preprocessor_dir:
      self.preprocessor_dir = os.path.normpath(os.path.join(self.build_dir, self.preprocessor_dir))
      if not os.path.exists(self.preprocessor_dir):
        os.makedirs(self.preprocessor_dir)
    self.preprocessor_no_o = preprocessor_no_o
    self.preprocessor_args = preprocessor_args
    if self.preprocessor:
      preprocessor_exist = False
      for path in os.environ["PATH"].split(os.pathsep):
        preprocessor_exist = os.path.exists(os.path.join(path, self.preprocessor))
        if preprocessor_exist:
          break
      if not preprocessor_exist:
        self.print_w('Error: ' + self.preprocessor + ' is not in your PATH! You cannot use -preprocessor switch.')
    return

  def _compile_preprocessor(self, file_to_compile):
    """
    Create compile command with preprocessor usage.

    Parameters
    ----------
    file_to_compile : ParsedFile object
      file to be compiled

    Returns
    -------
    str
      string containing the preprocessor command
    str
      string containing the output file name of preprocessor
    str
      string containing the command for removing/storing preprocessor outputs
    """
    preprocessor_cmd = ''
    preprocessor_output = ''
    preprocessor_remove = ''
    to_preprocess = False
    if self.preprocessor:
      if len(self.preprocessor_ext) > 0:
        if file_to_compile.extension in self.preprocessor_ext:
          to_preprocess = True
      else:
        to_preprocess = True
      if to_preprocess and self.preprocessor_dir:
        preprocessor_dir = self.preprocessor_dir
        preprocessor_store = True
      else:
        preprocessor_dir = ''
        preprocessor_store = False
      if to_preprocess:
        output_prepend = ' -o '
        if self.preprocessor_no_o:
          output_prepend = ' '
        preprocessor_cmd = self.preprocessor + ' ' + self.preprocessor_args + ' ' + file_to_compile.name + output_prepend + os.path.join(preprocessor_dir, file_to_compile.basename + '.pp.f90') + ' ; '
        preprocessor_output = os.path.join(preprocessor_dir, file_to_compile.basename + '.pp.f90')
        if not preprocessor_store:
          preprocessor_remove = ' ; rm -f ' + preprocessor_output
    return preprocessor_cmd, preprocessor_output, preprocessor_remove

  def _compile_include(self):
    """
    Method for creating compile command with inluded paths.

    Returns
    -------
    str
      string containing the include command
    """
    include_cmd = ''
    if len(self.dinc) > 0:
      include_cmd = ''.join(['-I' + s + ' ' for s in self.dinc])
    return include_cmd

  def _compile_command(self, file_to_compile):
    """
    Method for returning the OS command for compiling file_to_compile.

    Parameters
    ----------
    file_to_compile : ParsedFile object
      file to be compiled

    Returns
    -------
    str
      string containing the compile command
    """
    preprocessor_cmd, preprocessor_output, preprocessor_remove = self._compile_preprocessor(file_to_compile)
    include_cmd = self._compile_include()

    if preprocessor_cmd != '':
      comp_cmd = preprocessor_cmd + self.cmd_comp + ' ' + include_cmd + preprocessor_output + ' -o ' + os.path.join(self.obj_dir, file_to_compile.basename + '.o') + preprocessor_remove
    else:
      comp_cmd = self.cmd_comp + ' ' + include_cmd + file_to_compile.name + ' -o ' + os.path.join(self.obj_dir, file_to_compile.basename + '.o')
    return comp_cmd

  def _link_command(self, file_to_build, output=None, nomodlibs=None, submodules=None, mklib=None):
    """
    Return the OS command for linkng file_to_build.

    Parameters
    ----------
    file_to_build : ParsedFile object
      file to be built
    output : str
      output name
    nomodlibs : {None}
      list of old-Fortran style libraries objects
    submodules : {None}
      list of submodules objects
    mklib : {None}
      activate building library mode

    Returns
    -------
    str
      string containing the link command
    """
    exe = self.get_output_name(file_to_build=file_to_build, output=output)
    link_cmd = self.cmd_link + " " + self._get_libs_link_command(file_to_build=file_to_build, nomodlibs=nomodlibs, submodules=submodules, mklib=mklib) + " -o " + exe
    return link_cmd, exe

  def _mklib_command(self, file_to_build, output=None, nomodlibs=None, submodules=None, mklib=None):
    """
    Return the OS command for linkng file_to_build as a library.

    Parameters
    ----------
    file_to_build : ParsedFile object
      file to be built
    output : str
      output name
    nomodlibs : {None}
      list of old-Fortran style libraries objects
    submodules : {None}
      list of submodules objects
    mklib : {None}
      activate building library mode

    Returns
    -------
    str
      string containing the link command
    """
    lib = self.get_output_name(file_to_build=file_to_build, output=output, mklib=mklib)
    link_cmd = self._get_libs_link_command(file_to_build=file_to_build, exclude_programs=True, nomodlibs=nomodlibs, submodules=submodules, mklib=mklib)
    if mklib is not None:
      if mklib.lower() == 'shared':
        link_cmd = self.cmd_link + " " + link_cmd + " -o " + lib
      elif mklib.lower() == 'static':
        link_cmd = self.ar + " " + self.arflags + " " + lib + " " + link_cmd
        if self.ranlib:
          link_cmd += " \n " + self.ranlib + " " + lib
    return link_cmd, lib

  def _get_libs_link_command(self, file_to_build, exclude_programs=False, nomodlibs=None, submodules=None, mklib=None):
    """
    Return the libraries link command

    Parameters
    ----------
    file_to_build : ParsedFile object
      file to be built
    exclude_programs : {False}
      flag for excluding programs obj from the list
    nomodlibs : {None}
      list of old-Fortran style libraries objects
    submodules : {None}
      list of submodules objects
    mklib : {None}
      activate building library mode

    Returns
    -------
    str
      string containing the link command
    """
    objs = []
    if nomodlibs is not None:
      objs = objs + nomodlibs
    if submodules is not None:
      objs = objs + submodules
    objs = objs + file_to_build.obj_dependencies(exclude_programs=exclude_programs)
    objs = list(set(objs)) # remove duplicated objs
    link_cmd = "".join([os.path.join(self.obj_dir, s + " ") for s in objs]) + "".join([s + " " for s in self.libs]) + "".join([s + " " for s in self.vlibs])
    if mklib is None or mklib.lower() == 'shared':
      if len(self.ext_libs) > 0:
        link_cmd += " " + "".join(["-l" + s + " " for s in self.ext_libs])
      if len(self.ext_vlibs) > 0:
        link_cmd += " " + "".join(["-l" + s + " " for s in self.ext_vlibs])
      if len(self.lib_dir) > 0:
        link_cmd += " " + "".join(["-L" + s + " " for s in self.lib_dir])
    return link_cmd

  def _get_hierarchy(self, file_to_build):
    """
    Create a hierarchy of compiling commands accordingly to the dependencies order.

    Parameters
    ----------
    file_to_build : ParsedFile
    """
    order_max = max([p for p in file_to_build.pfile_dep_all if not p.include and p.to_compile], key=operator.attrgetter('order')).order + 1
    hierarchy = [[] for _ in range(order_max)]
    for dep in file_to_build.pfile_dep_all:
      if dep.to_compile and not dep.include:
        hierarchy[dep.order].append([dep.name, self._compile_command(file_to_compile=dep)])
        dep.to_compile = False
    hierarchy = [h for h in hierarchy if len(h) > 0]
    return hierarchy

  def _compile(self, file_to_build, verbose=False, log=False, quiet=False):
    """
    Compile file.

    Parameters
    ----------
    file_to_build : ParsedFile
    verbose : {False}
      bool for activate extreme verbose outputs
    log : {False}
      bool for activate errors log saving

    Returns
    -------
    build_ok : bool
      flag for checking the building is completed;
      it is used for purging out failed non-module-library in order to try to build the target anywhere;
    """
    build_ok = True
    hierarchy = self._get_hierarchy(file_to_build)
    for deps in reversed(hierarchy):
      files_to_compile = ''
      cmds = []
      for dep in deps:
        files_to_compile = files_to_compile + " " + dep[0]
        cmds.append(dep[1])
      if len(deps) > 1 and self.jobs > 1 and __parallel__:
        jobs = min(len(deps), self.jobs)
        if not quiet:
          self.print_n("Compiling" + files_to_compile + " using " + str(jobs) + " concurrent processes")
          if verbose:
            self.print_n("Executing: " + str(cmds))
        pool = Pool(processes=jobs)
        results = pool.map(syswork, cmds)
        pool.close()
        pool.join()
      else:
        if not quiet:
          self.print_n("Compiling" + files_to_compile + " serially")
        results = []
        for cmd in cmds:
          if not quiet:
            if verbose:
              self.print_n("Executing: " + str(cmd))
          result = syswork(cmd)
          results.append(result)
      if not file_to_build.nomodlib:
        if log:
          check_results(results=results, log="building-errors.log", print_w=self.print_w)
        else:
          check_results(results=results, print_w=self.print_w)
    if file_to_build.nomodlib:
      if results[0][0] != 0:
        build_ok = False
        if not quiet:
          self.print_w('Warning: compiling file ' + file_to_build.name + ' fails! Removing from the list of non-module libraries linked into the target!')
    return build_ok

  def gnu_make(self):
    """
    Return the builder options formated as GNU Make variables

    Returns
    -------
    str
      string containing the builder options in GNU Make format
    """
    string = []
    string.append("DOBJ    = " + os.path.normpath(self.obj_dir) + "/\n")
    string.append("DMOD    = " + os.path.normpath(self.mod_dir) + "/\n")
    string.append("DEXE    = " + os.path.normpath(self.build_dir) + "/\n")
    libs = []
    if len(self.libs) > 0:
      libs += self.libs
    if len(self.vlibs) > 0:
      libs += self.vlibs
    if len(self.ext_libs) > 0:
      libs += ["-l" + l + " " for l in self.ext_libs]
    if len(self.ext_vlibs) > 0:
      libs += ["-l" + l + " " for l in self.ext_vlibs]
    if len(self.lib_dir) > 0:
      libs += ["-L" + l + " " for l in self.lib_dir]
    if len(libs) > 0:
      string.append("LIBS    = " + "".join(" " + l for l in libs) + "\n")
    else:
      string.append("LIBS    =\n")
    string.append("FC      = " + self.compiler.fcs + "\n")
    string.append("OPTSC   = " + self.compiler.cflags + " " + self.compiler.modsw + self.mod_dir + "\n")
    string.append("OPTSL   = " + self.compiler.lflags + " " + self.compiler.modsw + self.mod_dir + "\n")
    return "".join(string)

  def build(self, file_to_build, output=None, nomodlibs=None, submodules=None, mklib=None, verbose=False, log=False, quiet=False, track=False):
    """
    Build a file.

    Parameters
    ----------
    file_to_build : ParsedFile
    output : str
      output name
    nomodlibs : {None}
      list of old-Fortran style libraries objects
    submodules : {None}
      list of submodules objects
    mklib : {None}
      bool for activate building library mode
    verbose : {False}
      bool for activate extreme verbose outputs
    log : {False}
      bool for activate errors log saving
    quiet : {False}
      bool for activate quiet mode
    track : {False}
      bool for activate build traking mode for subsequent install command

    Returns
    -------
    build_ok : bool
      flag for checking the building is completed;
      it is used for purging out failed non-module-library in order to try to build the target anywhere;
    """
    build_ok = True
    if not quiet:
      self.print_n('Building ' + file_to_build.name)
    # sort dependencies accordingly to indirect dependency
    file_to_build.sort_dependencies()
    # creating a hierarchy list of compiling commands accordingly to the order of all dependencies
    if len([p for p in file_to_build.pfile_dep_all if not p.include and p.to_compile]) > 0:
      build_ok = self._compile(file_to_build=file_to_build, verbose=verbose, log=log, quiet=quiet)
    else:
      if not quiet:
        self.print_n('Nothing to compile, all objects are up-to-date')
    if file_to_build.program:
      link_cmd, exe = self._link_command(file_to_build=file_to_build, output=output, nomodlibs=nomodlibs, submodules=submodules, mklib=mklib)
      if not quiet:
        self.print_n("Linking " + exe)
        if verbose:
          self.print_n("Executing: " + str(link_cmd))
      result = syswork(link_cmd)
      if log:
        check_results(results=[result], log="building-errors.log", print_w=self.print_w)
      else:
        check_results(results=[result], print_w=self.print_w)
      if not quiet:
        self.print_n('Target ' + file_to_build.name + ' has been successfully built')
    elif mklib:
      link_cmd, lib = self._mklib_command(file_to_build=file_to_build, output=output, nomodlibs=nomodlibs, submodules=submodules, mklib=mklib)
      if not quiet:
        self.print_n("Linking " + lib)
        if verbose:
          self.print_n("Executing: " + str(link_cmd))
      result = syswork(link_cmd)
      if log:
        check_results(results=[result], log="building-errors.log", print_w=self.print_w)
      else:
        check_results(results=[result], print_w=self.print_w)
      if not quiet:
        self.print_n('Target ' + file_to_build.name + ' has been successfully built')
    if track:
      print('Track building of ' + file_to_build.name)
      track_file_name = self.get_track_build_file(file_to_build)
      with open(track_file_name, 'w') as track_file:
        track_file.writelines('[build]\n')
        track_file.writelines('output = ' + self.get_output_name(file_to_build=file_to_build, output=output) + '\n')
        if file_to_build.program:
          track_file.writelines('program = True\n')
        elif mklib:
          track_file.writelines('library = True\n')
          track_file.writelines('mod_file = ' + os.path.join(self.mod_dir, file_to_build.basename + '.mod') + '\n')
    return build_ok

  def get_output_name(self, file_to_build, output=None, mklib=None):
    """
    Return the output build file name.

    Parameters
    ----------
    file_to_build : ParsedFile
    output : str
      output build file name
    mklib : {None}
      activate building library mode

    Returns
    -------
    output : str
      output build file name
    """
    if output:
      build_name = os.path.join(self.build_dir, output)
    else:
      if mklib is not None:
        if mklib.lower() == 'shared':
          build_name = os.path.join(self.build_dir, file_to_build.basename + '.so')
        elif mklib.lower() == 'static':
          build_name = os.path.join(self.build_dir, file_to_build.basename + '.a')
      else:
        build_name = os.path.join(self.build_dir, file_to_build.basename)
    return build_name

  def get_track_build_file(self, file_to_build):
    """
    Return the file name of the 'track build' file.

    Parameters
    ----------
    file_to_build : ParsedFile

    Returns
    -------
    track_file_name : str
      file name of 'track build' file
    """
    return os.path.join(self.build_dir, '.' + os.path.basename(file_to_build.name) + '.track_build')

  def verbose(self, quiet=False):
    """
    The method verbose returns a verbose message containing builder infos.

    Parameters
    ----------
    quiet : {False}
      flag for making less verbose outputs
    """
    message = ''
    if not quiet:
      message = "Builder options" + "\n"
      message += "  Directories\n"
      message += '    Root of source files directory: "' + self.src_dir + '"\n'
      message += '    Building directory: "' + self.build_dir + '"\n'
      message += '    Compiled-objects .o   directory: "' + self.obj_dir + '"\n'
      message += '    Compiled-objects .mod directory: "' + self.mod_dir + '"\n'
      if len(self.lib_dir) > 0:
        message += "  External libraries directories: " + "".join([s + " " for s in self.lib_dir]) + "\n"
      if len(self.dinc) > 0:
        message += "  Included paths: " + "".join([s + " " for s in self.dinc]) + "\n"
      if len(self.libs) > 0:
        message += "  Linked libraries with full path: " + "".join([s + " " for s in self.libs]) + "\n"
      if len(self.vlibs) > 0:
        message += "  Linked volatile libraries with full path: " + "".join([s + " " for s in self.vlibs]) + "\n"
      if len(self.ext_libs) > 0:
        message += "  Linked libraries in path: " + "".join([s + " " for s in self.ext_libs]) + "\n"
      if len(self.ext_vlibs) > 0:
        message += "  Linked volatile libraries in path: " + "".join([s + " " for s in self.ext_vlibs]) + "\n"
      message += self.compiler.pprint(prefix='  ')
      message += "  Archiver: " + self.ar + " " + self.arflags + "\n"
      if self.ranlib:
        message += "  Ranlib: " + self.ranlib + "\n"
      message += "  Preprocessor used: " + str(self.preprocessor) + "\n"
      message += "  Preprocessor args: " + str(self.preprocessor_args) + "\n"
      message += "  Preprocessor output directory: " + str(self.preprocessor_dir) + "\n"
      message += "  Preprocessor extensions processed: " + str(self.preprocessor_ext) + "\n"
    return message



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File: src/main/python/fobis/Compiler.py
# Author: Stefano Zaghi <stefano.zaghi@gmail.com>
# Date: 01.08.2017
# Last Modified Date: 01.08.2017
# Last Modified By: Stefano Zaghi <stefano.zaghi@gmail.com>
"""
Compiler.py, module definition of Compiler class.
This is a class designed for handling compilers default support.
"""
# from __future__ import absolute_import
# from __future__ import division
# from __future__ import print_function
# from __future__ import unicode_literals
# Copyright (C) 2015  Stefano Zaghi
#
# This file is part of FoBiS.py.
#
# FoBiS.py is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# FoBiS.py is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with FoBiS.py. If not, see <http://www.gnu.org/licenses/>.
# from future import standard_library
# standard_library.install_aliases()
# from builtins import str
# from builtins import *
# from builtins import object
import re
from .utils import print_fake
__regex_opts__ = re.compile(r"-O[0-9,s]")


class Compiler(object):
  """
  Compiler is an object that handles the compilers default support, its attributes and methods.

  Attributes
  ----------
  supported : {['gnu', 'intel', 'intel_nextgen', 'g95', 'opencoarrays-gnu', 'pgi', 'ibm', 'nag', 'nvfortran', 'amd', 'custom']}
    list of supported compilers
  """

  supported = ['gnu', 'intel', 'intel_nextgen', 'g95', 'opencoarrays-gnu', 'pgi', 'ibm', 'nag', 'nvfortran', 'amd', 'custom']

  def __init__(self, cliargs, print_w=None):
    """
    Parameters
    ----------
    cliargs : argparse object
    print_w : {None}
      function for printing emphized warning message

    Attributes
    ----------
    compiler : {None}
      str containing compiler vendor name
    fcs : {None}
      str containing compiler statement
    cflags : {None}
      str containing compiling flags
    lflags : {None}
      str containing linking flags
    preproc : {None}
      str containing preprocessing flags
    modsw : {None}
      str containing compiler switch for modules searching path
    mpi : {False}
      activate the MPI compiler
    openmp : {False}
      activate the OpenMP pragmas
    openmp_offload : {False}
      activate the OpenMP Offload pragmas
    coarray : {False}
      activate the coarray compilation
    coverage : {False}
      activate the coverage instruments
    profile : {False}
      activate the profile instruments
    print_w : {None}
      function for printing emphized warning message
    """
    if print_w is None:
      self.print_w = print_fake
    else:
      self.print_w = print_w

    self._mpi = None
    self._openmp = None
    self._openmp_offload = None
    self._coarray = None
    self._coverage = None
    self._profile = None
    self.compiler = cliargs.compiler
    if self.compiler:
      if self.compiler.lower() == 'gnu':
        self._gnu()
      elif self.compiler.lower() == 'intel':
        self._intel()
      elif self.compiler.lower() == 'intel_nextgen':
        self._intel_nextgen()
      elif self.compiler.lower() == 'g95':
        self._g95()
      elif self.compiler.lower() == 'opencoarrays-gnu':
        self._opencoarrays_gnu()
      elif self.compiler.lower() == 'pgi':
        self._pgi()
      elif self.compiler.lower() == 'ibm':
        self._ibm()
      elif self.compiler.lower() == 'nag':
        self._nag()
      elif self.compiler.lower() == 'nvfortran':
        self._nvfortran()
      elif self.compiler.lower() == 'amd':
        self._amd()
      elif self.compiler.lower() == 'custom':
        self._custom()
      else:
        self._gnu()
    # overriding default values if passed
    if cliargs.fc:
      self.fcs = cliargs.fc
    if cliargs.cflags:
      self.cflags = cliargs.cflags
    if cliargs.lflags:
      self.lflags = cliargs.lflags
    if cliargs.preproc:
      self.preproc = cliargs.preproc
    if cliargs.modsw:
      self.modsw = cliargs.modsw.replace("'","")
    self.mpi = cliargs.mpi
    self.openmp = cliargs.openmp
    self.openmp_offload = cliargs.openmp_offload
    self.coarray = cliargs.coarray
    self.coverage = cliargs.coverage
    self.profile = cliargs.profile
    self._set_fcs()
    self._set_cflags()
    self._set_lflags(mklib=cliargs.mklib)
    return

  def __str__(self):
    return self.pprint()

  def _gnu(self):
    """Set compiler defaults to the GNU gfortran compiler options."""
    self.compiler = 'gnu'
    self.fcs = 'gfortran'
    self.cflags = '-c'
    self.lflags = ''
    self.preproc = ''
    self.modsw = '-J '
    self._mpi = 'mpif90'
    self._openmp = ['-fopenmp', '-fopenmp']
    self._coarray = ['-fcoarray=lib', '-fcoarray=lib -lcaf_mpi']
    self._coverage = ['-ftest-coverage -fprofile-arcs', '-fprofile-arcs']
    self._profile = ['-pg', '-pg']
    return

  def _intel(self):
    """Set compiler defaults to the Intel Fortran compiler options."""
    self.compiler = 'intel'
    self.fcs = 'ifort'
    self.cflags = '-c'
    self.lflags = ''
    self.preproc = ''
    self.modsw = '-module '
    self._mpi = 'mpiifort'
    self._openmp = ['-qopenmp', '-qopenmp']
    self._coarray = ['-coarray', '-coarray']
    self._coverage = ['-prof-gen=srcpos', '']
    self._profile = ['', '']
    return

  def _intel_nextgen(self):
    """Set compiler defaults to the Intel Next Generation Fortran compiler options."""
    self.compiler = 'intel_nextgen'
    self.fcs = 'ifx'
    self.cflags = '-c'
    self.lflags = ''
    self.preproc = ''
    self.modsw = '-module '
    self._mpi = 'mpiifort -fc=ifx'
    self._openmp = ['-qopenmp', '-fiopenmp']
    self._openmp_offload = '-fopenmp-targets=spir64'
    self._coarray = ['-coarray', '-coarray']
    self._coverage = ['-prof-gen=srcpos', '']
    self._profile = ['', '']
    return

  def _g95(self):
    """Set compiler defaults to the g95 compiler options."""
    self.compiler = 'g95'
    self.fcs = 'g95'
    self.cflags = '-c'
    self.lflags = ''
    self.preproc = ''
    self.modsw = '-fmod='
    self._mpi = 'mpif90'
    self._openmp = ['', '']
    self._coarray = ['', '']
    self._coverage = ['', '']
    self._profile = ['', '']
    return

  def _opencoarrays_gnu(self):
    """Set compiler defaults to the OpenCoarrays-Aware GNU gfortran compiler options."""
    self.compiler = 'opencoarrays-gnu'
    self.fcs = 'caf'
    self.cflags = '-c'
    self.lflags = ''
    self.preproc = ''
    self.modsw = '-J '
    self._mpi = 'mpif90'
    self._openmp = ['-fopenmp', '-fopenmp']
    self._coarray = ['', '']
    self._coverage = ['-ftest-coverage -fprofile-arcs', '-fprofile-arcs']
    self._profile = ['-pg', '-pg']
    return

  def _pgi(self):
    """Set compiler defaults to the PGI compiler options."""
    self.compiler = 'pgi'
    self.fcs = 'pgfortran'
    self.cflags = '-c'
    self.lflags = ''
    self.preproc = ''
    self.modsw = '-module '
    self._mpi = 'mpif90'
    self._openmp = ['-mp', '-mp']
    self._coarray = ['', '']
    self._coverage = ['', '']
    self._profile = ['-pg', '-pg']
    return

  def _ibm(self):
    """Set compiler defaults to the IBM XL Fortran compiler options."""
    self.compiler = 'ibm'
    self.fcs = 'xlf2008_r'
    self.cflags = '-c'
    self.lflags = ''
    self.preproc = ''
    self.modsw = '-qmoddir='
    self._mpi = 'mpif90'
    self._openmp = ['-qsmp=omp', '-qsmp=omp']
    self._coarray = ['', '']
    self._coverage = ['', '']
    self._profile = ['-pg', '-pg']
    return

  def _nag(self):
    """Set compiler defaults to the NAG fortran compiler options."""
    self.compiler = 'gnu'
    self.fcs = 'nagfor'
    self.cflags = '-c'
    self.lflags = ''
    self.preproc = ''
    self.modsw = '-mdir '
    self._mpi = 'mpif90'
    self._openmp = ['-openmp', '-openmp']
    self._coarray = ['-num_images=1', '']
    self._coverage = ['', '']
    self._profile = ['-pg', '-pg']
    return

  def _nvfortran(self):
    """Set compiler defaults to the NVFortran compiler options."""
    self.compiler = 'nvfortran'
    self.fcs = 'nvfortran'
    self.cflags = '-c'
    self.lflags = ''
    self.preproc = ''
    self.modsw = '-module '
    self._mpi = 'mpif90'
    self._openmp = ['-mp', '-mp']
    self._coarray = ['', '']
    self._coverage = ['', '']
    self._profile = ['-pg', '-pg']
    return

  def _amd(self):
    """Set compiler defaults to the AMD Flang compiler options."""
    self.compiler = 'amd'
    self.fcs = 'amdflang'
    self.cflags = '-c'
    self.lflags = ''
    self.preproc = ''
    self.modsw = '-module-dir '
    self._mpi = 'mpif90'
    self._openmp = ['-fopenmp', '-fopenmp']
    self._coarray = ['', '']
    self._coverage = ['', '']
    self._profile = ['-pg', '-pg']
    return

  def _custom(self):
    """Set compiler defaults to be empty."""
    self.compiler = ''
    self.fcs = ''
    self.cflags = ''
    self.lflags = ''
    self.preproc = ''
    self.modsw = ''
    self._mpi = ''
    self._openmp = ''
    self._coarray = ''
    self._coverage = ''
    self._profile = ''
    return

  def _set_fcs(self):
    """Set the compiler command statement directly depending on the compiler."""
    if self.compiler.lower() in Compiler.supported:
      if self.mpi:
        self.fcs = self._mpi
    return

  def _set_cflags(self):
    """Set the compiling flags directly depending on the compiler."""
    if self.coverage:
      if self._coverage[0] != '':
        if re.search(__regex_opts__, self.cflags):
          self.print_w('Warning: found optimizations cflags within coverage ones: coverage results can be alterated!')
        self.cflags += ' -Og ' + self._coverage[0]
    if self.profile:
      if self._profile[0] != '':
        self.cflags += ' ' + self._profile[0]
    if self.openmp:
      if self._openmp[0] != '':
        self.cflags += ' ' + self._openmp[0]
    if self.openmp_offload:
      if self._openmp_offload != '':
        self.cflags += ' ' + self._openmp_offload
    if self.coarray:
      if self._coarray[0] != '':
        self.cflags += ' ' + self._coarray[0]
    if self.preproc is not None:
      if self.preproc != '':
        self.cflags += ' ' + self.preproc
    self.cflags = re.sub(r" +", r" ", self.cflags)
    return

  def _set_lflags(self, mklib=None):
    """Set the linking flags directly depending on the compiler."""
    if mklib is not None:
      if mklib.lower() == 'shared':
        if '-shared' not in self.lflags:
          self.lflags += ' -shared'
    if self.coverage:
      if self._coverage[1] != '':
        if re.search(__regex_opts__, self.lflags):
          self.print_w('Warning: found optimizations lflags within coverage ones: coverage results can be alterated!')
        self.lflags += ' -Og ' + self._coverage[1]
    if self.profile:
      if self._profile[1] != '':
        self.lflags += ' ' + self._profile[1]
    if self.openmp:
      if self._openmp[1] != '':
        self.lflags += ' ' + self._openmp[1]
    if self.openmp_offload:
      if self._openmp_offload != '':
        self.lflags += ' ' + self._openmp_offload
    if self.coarray:
      if self._coarray[1] != '':
        self.lflags += ' ' + self._coarray[1]
    self.lflags = re.sub(r" +", r" ", self.lflags)
    return

  def compile_cmd(self, mod_dir):
    """
    Method returning the compile command accordingly to the compiler options.

    Parameters
    ----------
    mod_dir : str
      path of the modules directory
    """
    return self.fcs + ' ' + self.cflags + ' ' + self.modsw + mod_dir + ' -I' + mod_dir

  def link_cmd(self, mod_dir):
    """
    Method returning the compile command accordingly to the compiler options.

    Parameters
    ----------
    mod_dir : str
      path of the modules directory
    """
    return self.fcs + ' ' + self.lflags + ' ' + self.modsw + mod_dir

  def pprint(self, prefix=''):
    """
    Pretty printer.

    Parameters
    ----------
    prefix : {''}
      prefixing string of each line
    """
    string = prefix + 'Compiler options\n'
    string += prefix + '  Vendor: "' + self.compiler.strip() + '"\n'
    string += prefix + '  Compiler command: "' + self.fcs.strip() + '"\n'
    string += prefix + '  Module directory switch: "' + self.modsw.strip() + '"\n'
    string += prefix + '  Compiling flags: "' + self.cflags.strip() + '"\n'
    string += prefix + '  Linking flags: "' + self.lflags.strip() + '"\n'
    string += prefix + '  Preprocessing flags: "' + self.preproc.strip() + '"\n'
    string += prefix + '  Coverage: ' + str(self.coverage) + '\n'
    if self.coverage:
      string += prefix + '    Coverage compile and link flags: ' + str(self._coverage) + '\n'
    string += prefix + '  Profile: ' + str(self.profile) + '\n'
    if self.profile:
      string += prefix + '    Profile compile and link flags: ' + str(self._profile) + '\n'
    return string



"""
ParsedFile.py, module definition of Dependency class.
This is a class designed for handling a single parsed file.
"""
# from __future__ import absolute_import
# from __future__ import division
# from __future__ import print_function
# from __future__ import unicode_literals
# Copyright (C) 2015  Stefano Zaghi
#
# This file is part of FoBiS.py.
#
# FoBiS.py is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# FoBiS.py is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with FoBiS.py. If not, see <http://www.gnu.org/licenses/>.

# from future import standard_library
# standard_library.install_aliases()
# from builtins import str
# from builtins import *
# from builtins import object
import sys

try:
  import graphviz as gvz
  import functools
  __graph__ = functools.partial(gvz.Graph, format='svg')
  __digraph__ = functools.partial(gvz.Digraph, format='svg')
  __styles__ = {'graph': {'fontsize': '16',
                          'fontcolor': 'black',
                          'bgcolor': 'white',
                          'rankdir': 'BT'},
                'nodes': {'fontname': 'Helvetica',
                          'shape': 'box',
                          'fontcolor': 'black',
                          'color': 'black',
                          'style': 'filled, rounded',
                          'fillcolor': 'white'},
                'edges': {'style': 'dashed',
                          'color': 'black',
                          'arrowhead': 'open',
                          'fontname': 'Courier',
                          'fontsize': '12',
                          'fontcolor': 'black'}}
  __graphviz__ = True

  def add_nodes(graph, nodes):
    """
    Auxiliary function for adding nodes to the dependency graph.

    Parameters
    ----------
    graph : __graph__ object
    nodes : nodes to be added to the graph
    """
    for node in nodes:
      if isinstance(node, tuple):
        graph.node(node[0], **node[1])
      else:
        graph.node(node)
    return graph

  def add_edges(graph, edges):
    """
    Auxiliary function for adding edges to the dependency graph.

    Parameters
    ----------
    graph : __graph__ object
    edges : nodes to be added to the graph
    """
    for edge in edges:
      if isinstance(edge[0], tuple):
        graph.edge(*edge[0], **edge[1])
      else:
        graph.edge(*edge)
    return graph

  def apply_styles(graph):
    """
    Auxiliary function for adding styles to the dependency graph.

    Parameters
    ----------
    graph : __graph__ object
    """
    graph.graph_attr.update(('graph' in __styles__ and __styles__['graph']) or {})
    graph.node_attr.update(('nodes' in __styles__ and __styles__['nodes']) or {})
    graph.edge_attr.update(('edges' in __styles__ and __styles__['edges']) or {})
    return graph
except ImportError:
  __graphviz__ = False
import operator
import os
import re
from subprocess import check_output, STDOUT
from .Dependency import Dependency
from .Doctest import Doctest
from .utils import traverse_recursive
from .utils import unique_seq
# definition of regular expressions
__str_apex__ = r"('|" + r'")'
__str_kw_include__ = r"[Ii][Nn][Cc][Ll][Uu][Dd][Ee]"
__str_kw_intrinsic__ = r"[Ii][Nn][Tt][Rr][Ii][Nn][Ss][Ii][Cc]"
__str_kw_iso_fortran_env__ = r"[Ii][Ss][Oo]_[Ff][Oo][Rr][Tt][Rr][Aa][Nn]_[Ee][Nn][Vv]"
__str_kw_iso_c_binding__ = r"[Ii][Ss][Oo]_[Cc]_[Bb][Ii][Nn][Dd][Ii][Nn][Gg]"
__str_kw_ieee_exceptions__ = r"[Ii][Ee][Ee][Ee]_[Ee][Xx][Cc][Ee][Pp][Tt][Ii][Oo][Nn][Ss]"
__str_kw_ieee_arithmetic__ = r"[Ii][Ee][Ee][Ee]_[Aa][Rr][Ii][Tt][Hh][Mm][Ee][Tt][Ii][Cc]"
__str_kw_ieee_features__ = r"[Ii][Ee][Ee][Ee]_[Ff][Ee][Aa][Tt][Uu][Rr][Ee][Ss]"
__str_kw_openacc__ = r"[Oo][Pp][Ee][Nn][Aa][Cc][Cc]"
__str_kw_omp_lib__ = r"[Oo][Mm][Pp]_[Ll][Ii][Bb]"
__str_kw_module__ = r"[Mm][Oo][Dd][Uu][Ll][Ee]"
__str_kw_submodule__ = r"[Ss][Uu][Bb][Mm][Oo][Dd][Uu][Ll][Ee]"
__str_kw_program__ = r"[Pp][Rr][Oo][Gg][Rr][Aa][Mm]"
__str_kw_use__ = r"[Uu][Ss][Ee]"
__str_kw_mpifh__ = r"[Mm][Pp][Ii][Ff]\.[Hh]"
__str_name__ = r"(?P<name>[a-zA-Z][a-zA-Z0-9_]*)"
__str_submodule_ancestor__ = r"(\((?P<submancestor>[a-zA-Z][a-zA-Z0-9_]*)\))"
__str_eol__ = r"(?P<eol>\s*!.*|\s*)?$"
# __str_f95_mod_rename__ = r"(\s*,\s*[a-zA-Z][a-zA-Z0-9_]*\s*=>\s*[a-zA-Z][a-zA-Z0-9_]*)*"
# __str_f95_mod_only__ = r"(\s*,\s*[Oo][Nn][Ll][Yy]\s*:\s*([a-zA-Z][a-zA-Z0-9_]*\s*=>\s*[a-zA-Z][a-zA-Z0-9_]*|[a-zA-Z][a-zA-Z0-9_]*))*"
__str_use_mod__ = (r"^(\s*)" +  # eventual initial white spaces
                   __str_kw_use__ +  # keyword "use"
                   # r"(\s*,\s*.*\s*::)*" +  # eventual module nature
                   # r"(\s+)" +  # 1 or more white spaces
                   r"((\s*,\s*.*\s*::)|(\s*::\s*)|(\s+))" + # eventual module nature
                   __str_name__ +  # construct name
                   r"(?P<mod_eol>(.*))")
__str_use_mod_intrinsic__ = (r"^(\s*)" +  # eventual initial white spaces
                             __str_kw_use__ +  # keyword "use"
                             r"\s*,\s*" + __str_kw_intrinsic__ + r"\s*::.*" +  # keyword intrinsic
                             r"(?P<mod_eol>(.*))")
__str_use_mod_iso_fortran_env__ = (r"^(\s*)" +  # eventual initial white spaces
                                   __str_kw_use__ +  # keyword "use"
                                   r"\s+" + __str_kw_iso_fortran_env__ +  # keyword intrinsic module iso_fortran_env
                                   r"(?P<mod_eol>(.*))")
__str_use_mod_iso_c_binding__ = (r"^(\s*)" +  # eventual initial white spaces
                                 __str_kw_use__ +  # keyword "use"
                                 r"\s+" + __str_kw_iso_c_binding__ +  # keyword intrinsic module iso_c_binding
                                 r"(?P<mod_eol>(.*))")
__str_use_mod_ieee_exceptions__ = (r"^(\s*)" +  # eventual initial white spaces
                                   __str_kw_use__ +  # keyword "use"
                                   r"\s+" + __str_kw_ieee_exceptions__ +  # keyword intrinsic module ieee_exceptions
                                   r"(?P<mod_eol>(.*))")
__str_use_mod_ieee_arithmetic__ = (r"^(\s*)" +  # eventual initial white spaces
                                   __str_kw_use__ +  # keyword "use"
                                   r"\s+" + __str_kw_ieee_arithmetic__ +  # keyword intrinsic module ieee_arithmetic
                                   r"(?P<mod_eol>(.*))")
__str_use_mod_ieee_features__ = (r"^(\s*)" +  # eventual initial white spaces
                                 __str_kw_use__ +  # keyword "use"
                                 r"\s+" + __str_kw_ieee_features__ +  # keyword intrinsic module ieee_features
                                 r"(?P<mod_eol>(.*))")
__str_use_mod_openacc__ = (r"^(\s*)" +  # eventual initial white spaces
                           __str_kw_use__ +  # keyword "use"
                           r"\s+" + r"(.*)" + __str_kw_openacc__ + r".*" + # keyword intrinsic module openacc
                           r"(?P<mod_eol>(.*))")
__str_use_mod_omp_lib__ = (r"^(\s*)" +  # eventual initial white spaces
                           __str_kw_use__ +  # keyword "use"
                           r"\s+" + r"(.*)" + __str_kw_omp_lib__ + r".*" + # keyword intrinsic module omp_lib
                           r"(?P<mod_eol>(.*))")
__str_include__ = (r"^(\s*|\#)" +  # eventual initial white spaces or "#" character
                   __str_kw_include__ +  # keyword "include"
                   r"(\s+)" +  # 1 or more white spaces
                   __str_apex__ +  # character "'" or '"'
                   r"(\s*)" +  # eventaul white spaces
                   r"(?P<name>.*)" +  # name of included file
                   r"(\s*)" +  # eventaul white spaces
                   __str_apex__ +  # character "'" or '"'
                   __str_eol__)  # eventual eol white space and or comment
__str_module__ = (r"^(\s*)" +  # eventual initial white spaces
                  r"(?P<scplevel>" + __str_kw_module__ + r")" +  # keyword "module"
                  r"(\s+)" +  # 1 or more white spaces
                  __str_name__ +  # construct name
                  __str_eol__)  # eventual eol white space and or comment
__str_submodule__ = (r"^(\s*)" +  # eventual initial white spaces
                     r"(?P<scplevel>" + __str_kw_submodule__ + r")" +  # keyword "submodule"
                     r"(\s+)" +  # 1 or more white spaces
                     # r"(\(.*\))" +
                     __str_submodule_ancestor__ +  # (ancestor-module-name [:parent-submodule-name])
                     r"(\s+)" +  # 1 or more white spaces
                     __str_name__ +  # construct name
                     __str_eol__)  # eventual eol white space and or comment
__str_program__ = (r"^(\s*)" +  # eventual initial white spaces
                   r"(?P<scplevel>" + __str_kw_program__ + r")" +  # keyword "program"
                   r"(\s+)" +  # 1 or more white spaces
                   __str_name__ +  # construct name
                   __str_eol__)  # eventual eol white space and or comment
__str_mpifh__ = (r"(.*)" + __str_kw_mpifh__ + r"(.*)")
__regex_use_mod__ = re.compile(__str_use_mod__)
__regex_include__ = re.compile(__str_include__)
__regex_program__ = re.compile(__str_program__)
__regex_module__ = re.compile(__str_module__)
__regex_submodule__ = re.compile(__str_submodule__)
__regex_use_mod_intrinsic__ = re.compile(__str_use_mod_intrinsic__)
__regex_use_mod_iso_fortran_env__ = re.compile(__str_use_mod_iso_fortran_env__)
__regex_use_mod_iso_c_binding__ = re.compile(__str_use_mod_iso_c_binding__)
__regex_use_mod_ieee_exceptions__ = re.compile(__str_use_mod_ieee_exceptions__)
__regex_use_mod_ieee_arithmetic__ = re.compile(__str_use_mod_ieee_arithmetic__)
__regex_use_mod_ieee_features__ = re.compile(__str_use_mod_ieee_features__)
__regex_use_mod_openacc__ = re.compile(__str_use_mod_openacc__)
__regex_use_mod_omp_lib__ = re.compile(__str_use_mod_omp_lib__)
__regex_use_intrinsic_modules__ = [__regex_use_mod_intrinsic__,
                                   __regex_use_mod_iso_fortran_env__,
                                   __regex_use_mod_iso_c_binding__,
                                   __regex_use_mod_ieee_exceptions__,
                                   __regex_use_mod_ieee_arithmetic__,
                                   __regex_use_mod_ieee_features__,
                                   __regex_use_mod_openacc__,
                                   __regex_use_mod_omp_lib__]
__regex_mpifh__ = re.compile(__str_mpifh__)

# alternative "open and read"... it should avoid encoding issues with python 2.X vs 3.X
def openReader(filename):
  if sys.version_info[0] < 3:
    return open(filename, 'r')
  else:
    return open(filename, 'r', newline='', encoding='utf8')

class ParsedFile(object):
  """ParsedFile is an object that handles a single parsed file, its attributes and methods."""
  def __init__(self, name, program=False, module=False, submodule=False, include=False, nomodlib=False, to_compile=False, output=None, is_doctest=False):
    """
    Parameters
    ----------
    name : str
      file name
    program : {False}
      flag for tagging program file
    module : {False}
      flag for tagging module file
    submodule : {False}
      flag for tagging submodule file
    include : {False}
      flag for tagging include file
    nomodlib : {False}
      flag for tagging library of procedures without an enclosing module (old Fortran style)
    to_compile : {False}
      flag for tagging file to be compiled
    output : {None}
      string of output file name

    Attributes
    ----------
    name : str
      file name
    program : bool
      flag for tagging program file
    module : bool
      flag for tagging module file
    submodule : bool
      flag for tagging submodule file
    include : bool
      flag for tagging include file
    nomodlib : bool
      flag for tagging library of procedures without an enclosing module (old Fortran style)
    to_compile : bool
      flag for tagging file to be compiled
    output : str
      string of output file name
    basename : str
      basename of file name
    extension : str
      extension of file name
    timestamp : str
      timestamp of file
    order :
    pfile_dep : list
      list of parsed file dependencies
    doctest : Doctest()
      doctest object
    """
    self.name = name
    self.program = program
    self.module = module
    self.submodule = submodule
    self.include = include
    self.nomodlib = nomodlib
    self.to_compile = to_compile
    self.output = output
    self.is_doctest = is_doctest
    self.basename = os.path.splitext(os.path.basename(self.name))[0]
    self.extension = os.path.splitext(os.path.basename(self.name))[1]
    self.timestamp = os.path.getmtime(self.name)
    self.order = 0
    self.pfile_dep = None
    self.pfile_dep_all = None
    self.module_names = None
    self.submodule_names = None
    self.dependencies = None
    self.doctest = None
    return

  def sort_dependencies(self):
    """
    Sort dependencies.
    """
    for dep in self.pfile_dep_all:
      for other_dep in self.pfile_dep_all:
        if other_dep != dep:
          if dep in other_dep.pfile_dep_all:
            dep.order += 1
    self.pfile_dep_all.sort(key=operator.attrgetter('order'), reverse=True)
    return

  def parse(self, inc=['.INC', '.F', '.FOR', '.FPP', '.F77', '.F90', '.F95', '.F03', '.F08'], preprocessor='cpp', preproc='', include=''):
    """
    Parse the file creating its the dependencies list and the list of modules names that self eventually contains.

    Parameters
    ----------
    inc : list
      list of extensions of included files
    preprocessor : str
      preprocessor name
    preproc : str
      preprocessor flags
    include : list
      include directories
    """
    self.module_names = []
    self.submodule_names = []
    self.dependencies = []

    if self.extension in inc:
       preprocessor_exist = False
       for path in os.environ["PATH"].split(os.pathsep):
          preprocessor_exist = os.path.exists(os.path.join(path, preprocessor))
          if preprocessor_exist:
             break
       if preprocessor_exist:
          if preprocessor == 'cpp':
             preprocessor += ' -C -w '
          elif preprocessor == 'fpp':
             preprocessor += ' -w '
          if preproc is None:
             preproc = ''
          includes = ' '
          if len(include) > 0 :
             includes = ' -I' + '-I'.join(include) + ' '
          source = str(check_output(preprocessor + ' ' + preproc + includes + self.name, shell=True, stderr=STDOUT, encoding='UTF-8'))
          source = source.replace('\\n', '\n')
       else:
         source = str(openReader(self.name).read())
    else:
       source = str(openReader(self.name).read())

    for line in source.split('\n'):
      matching = re.match(__regex_program__, line)
      if matching:
        self.program = True
      matching = re.match(__regex_module__, line)
      if matching:
        self.module = True
        self.module_names.append(matching.group('name'))
      matching = re.match(__regex_submodule__, line)
      if matching:
        self.submodule = True
        self.submodule_names.append(matching.group('name'))
        dep = Dependency(dtype="module", name=matching.group('submancestor'))
        self.dependencies.append(dep)
      matching = re.match(__regex_use_mod__, line)
      if matching:
        if not any(re.match(regex, line) for regex in __regex_use_intrinsic_modules__):
          if matching.group('name').lower() != 'mpi' and matching.group('name').lower() != 'omp_lib':
            dep = Dependency(dtype="module", name=matching.group('name'))
            self.dependencies.append(dep)
      matching = re.match(__regex_include__, line)
      if matching:
        if not re.match(__regex_mpifh__, line):
          dep = Dependency(dtype="include", name=matching.group('name'))
          self.dependencies.append(dep)

    if self.module:
      self.doctest = Doctest()
      self.doctest.parse(source=source)
      self.doctest.make_volatile_programs()

    if not self.program and not self.module and not self.submodule:
      if os.path.splitext(os.path.basename(self.name))[1] not in inc:
        self.nomodlib = True

  def save_build_log(self, builder):
    """
    Save a log file containing information about the building options used.

    Parameters
    ----------
    builder : Builder object
      builder used for building self
    """
    with open(os.path.join(builder.build_dir, "build_" + self.basename + ".log"), "w") as log_file:
      log_file.writelines("Hierarchical dependencies list of: " + self.name + "\n")
      for dep in self.pfile_dep:
        log_file.writelines("  " + dep.name + "\n")
        log_file.writelines(dep.str_dependencies(pref="    "))
      log_file.writelines("Complete ordered dependencies list of: " + self.name + "\n")
      for dep in self.pfile_dep_all:
        log_file.writelines("  " + dep.name + "\n")
      log_file.writelines(builder.verbose())
    return

  def save_dep_graph(self):
    """
    Save dependency graph.
    """
    if __graphviz__:
      depgraph = __digraph__()
      depgraph = add_nodes(depgraph, [self.name])
      depgraph.graph_attr['label'] = 'Dependencies graph of ' + self.name
      for dep in self.pfile_dep:
        depgraph = add_nodes(depgraph, [dep.name])
        depgraph = add_edges(depgraph, [(self.name, dep.name)])
        for subdep in dep.str_dependencies().split("\n"):
          if subdep != '':
            depgraph = add_nodes(depgraph, [subdep])
            depgraph = add_edges(depgraph, [(dep.name, subdep)])
      depgraph = apply_styles(depgraph)
      depgraph.render("dependency_graph_" + self.basename)
    else:
      print("Module 'graphviz' not found: saving of dependency graph disabled")
    return

  def gnu_make_rule(self, builder):
    """
    Return the file compiling rule in GNU Make format

    Parameters
    ----------
    builder : Builder object

    Returns
    -------
    str
      string containing the compiling rule of the file
    """
    string = []
    target = self.basename.lower()
    if not self.include:
      objs = []
      for dep in self.pfile_dep:
        if not dep.include and os.path.splitext(os.path.basename(dep.name))[0].lower() != target:
          objs.append("\t$(DOBJ)" + os.path.splitext(os.path.basename(dep.name))[0].lower() + ".o")
      deps = [self.name]
      for dep in self.pfile_dep:
        if dep.include:
          deps.append(dep.name)
      if len(objs) > 0:
        string.append("$(DOBJ)" + target + ".o:" + "".join([" " + dep for dep in deps]) + " \\" + "\n")
        for obj in objs[:-1]:
          string.append(obj + " \\" + "\n")
        string.append(objs[-1] + "\n")
      else:
        string.append("$(DOBJ)" + target + ".o:" + "".join([" " + dep for dep in deps]) + "\n")
      string.append("\t@echo $(COTEXT)\n")
      string.append("\t@$(FC) $(OPTSC) " + ''.join(['-I' + i + ' ' for i in builder.dinc]) + " $< -o $@\n")
      string.append("\n")
    return "".join(string)

  def str_dependencies(self, pref=""):
    """
    Create a string containing the depencies files list.

    Parameters
    ----------
    pref : {""}
      prefixing string
    """
    str_dep = ""
    for dep in self.pfile_dep:
      str_dep += pref + dep.name + "\n"
    return str_dep

  def obj_dependencies(self, exclude_programs=False):
    """
    Create a list containing the dependencies object files list.

    Parameters
    ----------
    exclude_programs : {False}
      flag for excluding programs obj from the list
    """
    if exclude_programs:
      return [p.basename + ".o" for p in self.pfile_dep_all if not p.include and not p.program]
    else:
      return [p.basename + ".o" for p in self.pfile_dep_all if not p.include]

  def check_compile(self, obj_dir, force_compile=False):
    """
    Check if self must be compiled.

    Parameters
    ----------
    obj_dir : str
      directory where compiled objects are saved
    force_compile : {False}
      flag for forcing (re-)compiling of all dependency
    """
    if not self.include:
      # verifying if dependencies are up-to-date
      for dep in self.pfile_dep_all:
        if not dep.include:
          if force_compile:
            self.to_compile = True
          else:
            obj = os.path.join(obj_dir, dep.basename + ".o")
            # verifying if dep is up-to-date
            if os.path.exists(obj):
              if os.path.getmtime(obj) < dep.timestamp:
                # found a dependency object that is out-of-date, thus self (and dep) must be compiled
                self.to_compile = True
            else:
              # compiled object of a dependency is absent, thus self must be compiled
              self.to_compile = True
        else:
          # verifying if dep is newer than self
          if not os.path.exists(dep.name):
            print(" Attention: file " + dep.name + " does not exist, but it is a dependency of file " + self.name)
            sys.exit(1)
          else:
            # comparing the include dependency with the self-compiled-object if exist
            obj = os.path.join(obj_dir, self.basename + ".o")
            # verifying if dep is up-to-date
            if os.path.exists(obj):
              if os.path.getmtime(obj) < os.path.getmtime(dep.name):
                # found an include that is newer than self-compiled-object, thus self must be compiled
                self.to_compile = True
      # verifying if self is up-to-date
      if not self.to_compile:
        obj = os.path.join(obj_dir, self.basename + ".o")
        if os.path.exists(obj):
          if os.path.getmtime(obj) < self.timestamp:
            # the compiled object is out-of-date, thus self must be compiled
            self.to_compile = True
        else:
          # compiled object is absent, thus self must be compiled
          self.to_compile = True

  def create_pfile_dep_all(self):
    """
    Create a complete list of all dependencies direct and indirect.
    """
    self.pfile_dep_all = []
    for path in traverse_recursive(self):
      self.pfile_dep_all.append(path[-1])
    self.pfile_dep_all = unique_seq(self.pfile_dep_all)
    return



"""
Doctest.py, module definition of Doctest class.

This is a class designed for performing introspective tests
by means of special docstrings.
"""
# Copyright (C) 2015  Stefano Zaghi
#
# This file is part of FoBiS.py.
#
# FoBiS.py is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# FoBiS.py is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with FoBiS.py. If not, see <http://www.gnu.org/licenses/>.
# from __future__ import print_function
# from __future__ import absolute_import
# from __future__ import division
# from __future__ import unicode_literals
# from future import standard_library
# standard_library.install_aliases()
# from builtins import str
# from builtins import *
# from builtins import object
import os
import re
__remodule__ = re.compile(r"^(\s*)" +  # eventual initial white spaces
                          r"([Mm][Oo][Dd][Uu][Ll][Ee])" +  # keyword "module"
                          r"(\s+)" +  # 1 or more white spaces
                          r"(?P<name>[a-zA-Z][a-zA-Z0-9_]*)" +  # module name
                          r"(?P<eol>\s*!.*|\s*)?$")  # eventual eol white space and or comment
__redoctest__ = re.compile(r"(?P<doctest>\!.```fortran(?P<test>.*?)\!.```.*?\!\=\>(?P<result>.*?)\<\<\<)", re.DOTALL)


class Doctest(object):
  """Doctest is an object that handles introspective doc-tests for Fortran procedures,
  like the Python doctest module does for Python codes.

  The syntax of a FoBiS.py doctest for Fortran procedures is:
  !$```fortran
  !$ ! valid fortran code for testing the procedure
  !$ ! the doctest must end with printing the expected result to the standard output
  !$```
  !=> !expected results <<<

  Note that the character "$" can be any character, likely it should be the "docmark" you
  select for producing the API documentation by means of the great FORD autodocumentation
  Fortran tool.
  """

  def __init__(self):
    """
    Parameters
    ----------
    """
    self.to_test = False
    self.tests = []

  def parse(self, source):
    """Parse source string for doctest occurences.

    Parameters
    ----------
    source : str
      source string with eventual doctests
    """
    modules = []
    for line in source.split('\n'):
      match = re.match(__remodule__, line)
      if match:
        modules.append(match)
    if len(modules) > 0:
      for match in re.finditer(__redoctest__, source):
        lines = match.group('test').split('\n')
        lines = [line.strip()[2:] for line in lines if not line.strip() == '']
        module_name = None
        for module in modules:
          if module.start() < match.start():
            module_name = module.group('name')
        self.tests.append({'module': module_name, 'source': lines, 'result': match.group('result').strip()})

  def make_volatile_programs(self):
    """Make a 'volatile' program test for each of the parsed doctests."""
    if len(self.tests) > 0:
      self.to_test = True
      for test in self.tests:
        program_source = ["program volatile_doctest"]
        program_source.append("use " + test['module'])
        for line in test['source']:
          program_source.append(line)
        program_source.append("endprogram volatile_doctest")
        test['program_source'] = '\n'.join(program_source)

  def save_volatile_programs(self, build_dir):
    """Save a 'volatile' program test for each of the parsed doctests.

    Parameters
    ----------
    build_dir : str
      build directory where volatile programs are placed

    Returns
    -------
    doctest_root_dir : str
      directory where doctests are saved
    """
    if len(self.tests) > 0:
      doctest_root_dir = os.path.join(build_dir, 'doctests-src')
      for t, test in enumerate(self.tests):
        doctest_dir = os.path.join(doctest_root_dir, test['module'])
        if not os.path.exists(doctest_dir):
          os.makedirs(doctest_dir)
        doctest_file = os.path.join(doctest_dir, test['module'] + '-doctest-' + str(t + 1) + '.f90')
        with open(doctest_file, 'w') as doc:
          doc.writelines(test['program_source'])
        doctest_file = os.path.join(doctest_dir, test['module'] + '-doctest-' + str(t + 1) + '.result')
        with open(doctest_file, 'w') as doc:
          doc.writelines(test['result'])
      return doctest_root_dir



"""
Cleaner.py, module definition of Cleaner class.
This is a class designed for controlling the cleaning phase.
"""
# from __future__ import absolute_import
# from __future__ import division
# from __future__ import print_function
# from __future__ import unicode_literals
# Copyright (C) 2015  Stefano Zaghi
#
# This file is part of FoBiS.py.
#
# FoBiS.py is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# FoBiS.py is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with FoBiS.py. If not, see <http://www.gnu.org/licenses/>.
# from future import standard_library
# standard_library.install_aliases()
# from builtins import *
# from builtins import object
import os
from .utils import print_fake


class Cleaner(object):
  """
  Cleaner is an object for cleaning current project.
  """
  def __init__(self, cliargs, print_w=None):
    """
    Parameters
    ----------
    cliargs : argparse object
    print_w : {None}
      function for printing emphized warning message

    Attributes
    ----------
    build_dir : str
      directory containing built files
    obj_dir : str
      directory containing compiled object files
    mod_dir : str
      directory containing .mod files
    target : str
      target source to be built
    output : str
      name of the building output
    mklib : str
      flag for building a library instead of a program
    print_w : function
      function for printing emphized warning message
    """

    if print_w is None:
      self.print_w = print_fake
    else:
      self.print_w = print_w

    self._sanitize_dirs(build_dir=cliargs.build_dir, obj_dir=cliargs.obj_dir, mod_dir=cliargs.mod_dir)
    self._sanitize_files(target=cliargs.target, output=cliargs.output)
    self.mklib = cliargs.mklib
    return

  def _sanitize_dirs(self, build_dir, obj_dir, mod_dir):
    """
    Method for sanitizing directory paths.

    Parameters
    ----------
    build_dir : str
      directory containing built files
    obj_dir : str
      directory containing compiled object files
    mod_dir : str
      directory containing .mod files
    """
    self.build_dir = os.path.normpath(build_dir)
    self.obj_dir = os.path.normpath(os.path.join(build_dir, obj_dir))
    self.mod_dir = os.path.normpath(os.path.join(build_dir, mod_dir))
    return

  def _sanitize_files(self, target, output):
    """
    Method for sanitizing files paths.

    Parameters
    target : {None}
      target source to be built
    output : {None}
      name of the building output
    ----------
    """
    if target:
      self.target = os.path.normpath(target)
    else:
      self.target = target
    if output:
      self.output = os.path.normpath(output)
    else:
      self.output = output
    return

  def clean_mod(self):
    """
    Clean compiled (s)mod files.
    """
    if os.path.exists(self.mod_dir):
      self.print_w('Removing all *.(s)mod files into "' + self.mod_dir + '"')
      for root, _, files in os.walk(self.mod_dir):
        for filename in files:
          if os.path.splitext(os.path.basename(filename))[1] == '.mod':
            os.remove(os.path.join(root, filename))
          if os.path.splitext(os.path.basename(filename))[1] == '.smod':
            os.remove(os.path.join(root, filename))

  def clean_obj(self):
    """
    Method for cleaning compiled objects files.
    """
    if os.path.exists(self.obj_dir):
      self.print_w('Removing all *.o files into "' + self.obj_dir + '"')
      for root, _, files in os.walk(self.obj_dir):
        for filename in files:
          if os.path.splitext(os.path.basename(filename))[1] == '.o':
            os.remove(os.path.join(root, filename))

  def clean_target(self):
    """
    Function clean_target clean compiled targets.
    """
    if self.target:
      if self.output:
        exe = self.output
      else:
        if self.mklib:
          if self.mklib.lower() == 'static':
            exe = os.path.splitext(os.path.basename(self.target))[0] + '.a'
          elif self.mklib.lower() == 'shared':
            exe = os.path.splitext(os.path.basename(self.target))[0] + '.so'
        else:
          exe = os.path.splitext(os.path.basename(self.target))[0]
      if os.path.exists(os.path.join(self.build_dir, exe)):
        self.print_w('Removing ' + self.build_dir + exe)
        os.remove(os.path.join(self.build_dir, exe))
      if os.path.exists('build_' + os.path.splitext(os.path.basename(self.target))[0] + '.log'):
        self.print_w('Removing build_' + os.path.splitext(os.path.basename(self.target))[0] + '.log')
        os.remove('build_' + os.path.splitext(os.path.basename(self.target))[0] + '.log')
      if os.path.exists('dependency_graph_' + os.path.splitext(os.path.basename(self.target))[0] + '.svg'):
        self.print_w('Removing dependency_graph_' + os.path.splitext(os.path.basename(self.target))[0] + '.svg')
        os.remove('dependency_graph_' + os.path.splitext(os.path.basename(self.target))[0] + '.svg')
      if os.path.exists('dependency_graph_' + os.path.splitext(os.path.basename(self.target))[0]):
        self.print_w('Removing dependency_graph_' + os.path.splitext(os.path.basename(self.target))[0])
        os.remove('dependency_graph_' + os.path.splitext(os.path.basename(self.target))[0])
      if os.path.exists(os.path.join(self.build_dir, '.cflags.heritage')):
        os.remove(os.path.join(self.build_dir, '.cflags.heritage'))



#!/usr/bin/env python
"""Wrapper of FoBiS.py program extracted from fobis package"""
# Copyright (C) 2015  Stefano Zaghi
#
# This file is part of FoBiS.py.
#
# FoBiS.py is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# FoBiS.py is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with FoBiS.py. If not, see <http://www.gnu.org/licenses/>.
from fobis.fobis import main
import sys


if sys.platform == 'win32':
  import colorama
  colorama.init(convert=True)


if __name__ == '__main__':
  main()
